/*
 * User mode gadget driver to emulate a Gigabyte Odin GT power supply.
 *
 * Written in 2009 by Adam Nielsen
 *
 * Based on the gadgetfs usb.c sample application
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <memory.h>
#include <pthread.h>
#include <signal.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/poll.h>

#include <asm/byteorder.h>

#include <linux/types.h>
#include <linux/usb/gadgetfs.h>
#include <linux/usb/ch9.h>

#include "usbstring.h"

static int verbose = 1;


static int	HIGHSPEED;
static char	*DEVNAME;
static char	*EP_IN_NAME;

/* gadgetfs currently has no chunking (or O_DIRECT/zerocopy) support
 * to turn big requests into lots of smaller ones; so this is "small".
 */
#define	USB_BUFSIZE	(7 * 1024)

static enum usb_device_speed	current_speed;

static pthread_t	ep0;
static int retfd = -1;

// Taken from hid.h
struct hid_class_descriptor {
        __u8  bDescriptorType;
        __le16 wDescriptorLength;
} __attribute__ ((packed));

struct hid_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;
        __le16 bcdHID;
        __u8  bCountryCode;
        __u8  bNumDescriptors;

        struct hid_class_descriptor desc[1];
} __attribute__ ((packed));
// End hid.h copy

#define DRIVER_VENDOR_NUM       0x1044		/* Chu Yuen (Gigabyte) */
#define DRIVER_PRODUCT_NUM      0x4001		/* Odin GT */

#define	STRINGID_MFGR		1
#define	STRINGID_PRODUCT	2

static struct usb_string stringtab [] = {
	{ STRINGID_MFGR,	"GIGABYTE TECHNOLOGY", },
	{ STRINGID_PRODUCT,	"GBT 800W PSU", },
};

static struct usb_gadget_strings strings = {
	.language =	0x0409,		/* "en-us" */
	.strings =	stringtab,
};


static struct usb_device_descriptor
device_desc = {
	.bLength =		sizeof device_desc,
	.bDescriptorType =	USB_DT_DEVICE,

	.bcdUSB = __constant_cpu_to_le16 (0x0110),
	.bDeviceClass = USB_CLASS_PER_INTERFACE,
	.bDeviceSubClass =	0,
	.bDeviceProtocol =	0,
	// .bMaxPacketSize0 ... set by gadgetfs - should be 8 (odin) but is set to 64
	.idVendor = __constant_cpu_to_le16 (DRIVER_VENDOR_NUM),
	.idProduct = __constant_cpu_to_le16 (DRIVER_PRODUCT_NUM),
	//.bcdDevice = __constant_cpu_to_le16(0x0101),  // TODO: Only appears as 0x100?!
	.iManufacturer =	STRINGID_MFGR,
	.iProduct =		STRINGID_PRODUCT,
	.iSerialNumber = 0,//	STRINGID_SERIAL,
	.bNumConfigurations =	1,
};

#define	MAX_USB_POWER		100

#define	CONFIG_VALUE		1

static const struct usb_config_descriptor
config = {
	.bLength = sizeof config,
	.bDescriptorType =	USB_DT_CONFIG,

	/* must compute wTotalLength ... */
	.bNumInterfaces =	1,
	.bConfigurationValue =	CONFIG_VALUE,
	.iConfiguration =	0,//STRINGID_CONFIG,
	.bmAttributes = USB_CONFIG_ATT_ONE | USB_CONFIG_ATT_SELFPOWER, // only one that works
	//.bmAttributes = USB_CONFIG_ATT_ONE | USB_CONFIG_ATT_WAKEUP, // correct for device
	.bMaxPower =		(MAX_USB_POWER + 1) / 2,
};

static struct usb_interface_descriptor
source_sink_intf = {
	.bLength =		sizeof source_sink_intf,
	.bDescriptorType =	USB_DT_INTERFACE,

	.bInterfaceClass =	USB_CLASS_HID,
	.iInterface =	0,//	STRINGID_INTERFACE,
};

/* Full speed configurations are used for full-speed only devices as
 * well as dual-speed ones (the only kind with high speed support).
 */

static struct hid_descriptor hid_desc = {
	.bLength = sizeof(struct hid_descriptor),
	.bDescriptorType = USB_DT_CS_DEVICE,
	.bcdHID = __constant_cpu_to_le16 (0x0110),
	.bNumDescriptors = 1,
	.desc[0].bDescriptorType = 34,
	.desc[0].wDescriptorLength = 32,
};

const char cReportDescriptor1[] = {
	0x06, 0x00, 0xff, // Global Usage Page (null)
	0x09, 0x00, // Local Usage (null)
	0xa1, 0x01, // Main Collection (Application)
	0x09, 0x01, // Local Usage (null)
	0x15, 0x00, // Global Logical Minimum (0)
	0x25, 0x09, // Global Logical Maximum (9)
	0x75, 0x08, // Global Report Size (8)
	0x95, 0x08, // Global Report Count (8)
	0x81, 0x02, // Main Input (2) - bitfield
	0x09, 0x02, // Local Usage (2/null)
	0x15, 0x00, // Global Logical Minimum (0)
	0x25, 0x09, // Global Logical Maximum (0)
	0x75, 0x08, // Global Report Size (8)
	0x95, 0x05, // Global Report Count (5)
	0x91, 0x02, // Main Output (2) - bitfield
	0xc0        // Main End Collection
};

// Full speed (USB1.1)
static struct usb_endpoint_descriptor fs_source_desc = {
	.bLength = USB_DT_ENDPOINT_SIZE,
	.bDescriptorType = USB_DT_ENDPOINT,

	.bmAttributes = USB_ENDPOINT_XFER_INT,
	/* NOTE some controllers may need FS bulk max packet size
	 * to be smaller.  it would be a chip-specific option.
	 */
	.wMaxPacketSize = __constant_cpu_to_le16 (8),//64),
	.bInterval = 10, // 10ms period
};

// High speed (USB2.0) - we don't want this since the device is only USB1.1 but
// gadgetfs seems to require it.
static struct usb_endpoint_descriptor
hs_source_desc = {
	.bLength = USB_DT_ENDPOINT_SIZE,
	.bDescriptorType = USB_DT_ENDPOINT,

	.bmAttributes = USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize = __constant_cpu_to_le16 (8),//512),
	.bInterval = 7, // 8ms period
};

/* some devices can handle other status packet sizes */
#define STATUS_MAXPACKET	8
#define	LOG2_STATUS_POLL_MSEC	3

/*static struct usb_endpoint_descriptor
fs_status_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16 (STATUS_MAXPACKET),
	.bInterval =		(1 << LOG2_STATUS_POLL_MSEC),
};*/

static const struct usb_endpoint_descriptor *fs_eps [] = {
	&fs_source_desc,
//	&fs_sink_desc,
//	&fs_status_desc,
};


/* High speed configurations are used only in addition to a full-speed
 * ones ... since all high speed devices support full speed configs.
 * Of course, not all hardware supports high speed configurations.
 */


/*static struct usb_endpoint_descriptor
hs_status_desc = {
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16 (STATUS_MAXPACKET),
	.bInterval =		LOG2_STATUS_POLL_MSEC + 3,
};*/

static const struct usb_endpoint_descriptor *hs_eps [] = {
	&hs_source_desc,
//	&hs_sink_desc,
//	&hs_status_desc,
};


/*static inline int min(unsigned a, unsigned b)
{
	return (a < b) ? a : b;
}*/

static int autoconfig ()
{
	struct stat	statb;

	/* NetChip 2280 PCI device or dummy_hcd, high/full speed */
	if (stat (DEVNAME = "net2280", &statb) == 0 ||
			stat (DEVNAME = "dummy_udc", &statb) == 0) {
		HIGHSPEED = 1;
		device_desc.bMaxPacketSize0 = 8; // Doesn't work...
		device_desc.bcdDevice = __constant_cpu_to_le16 (0x0101);

		fs_source_desc.bEndpointAddress
			= hs_source_desc.bEndpointAddress
			= USB_DIR_IN | 1;
		EP_IN_NAME = "ep-a";
		//EP_IN_NAME = "ep10in-int";
//		fs_sink_desc.bEndpointAddress = hs_sink_desc.bEndpointAddress
//			= USB_DIR_OUT | 3;
//		EP_OUT_NAME = "ep-b";

		source_sink_intf.bNumEndpoints = 1;
		//fs_status_desc.bEndpointAddress
		//	= hs_status_desc.bEndpointAddress
		//	= USB_DIR_IN | 11;
		//EP_STATUS_NAME = "ep-f";

	} else {
		DEVNAME = 0;
		return -ENODEV;
	}
	return 0;
}



static void close_fd (void *fd_ptr)
{
	int	status, fd;

	fd = *(int *)fd_ptr;
	*(int *)fd_ptr = -1;

	/* test the FIFO ioctls (non-ep0 code paths) */
	if (pthread_self () != ep0) {
		status = ioctl (fd, GADGETFS_FIFO_STATUS);
		if (status < 0) {
			/* ENODEV reported after disconnect */
			if (errno != ENODEV && errno != -EOPNOTSUPP)
				perror ("get fifo status");
		} else {
			fprintf (stderr, "fd %d, unclaimed = %d\n",
				fd, status);
			if (status) {
				status = ioctl (fd, GADGETFS_FIFO_FLUSH);
				if (status < 0)
					perror ("fifo flush");
			}
		}
	}

	if (close (fd) < 0)
		perror ("close");
}


/* you should be able to open and configure endpoints
 * whether or not the host is connected
 */
static int
ep_config (char *name, const char *label,
	struct usb_endpoint_descriptor *fs,
	struct usb_endpoint_descriptor *hs
)
{
	int		fd, status;
	char		buf [USB_BUFSIZE];

	/* open and initialize with endpoint descriptor(s) */
	fd = open (name, O_RDWR);
	if (fd < 0) {
		status = -errno;
		fprintf (stderr, "%s open %s error %d (%s)\n",
			label, name, errno, strerror (errno));
		return status;
	}

	/* one (fs or ls) or two (fs + hs) sets of config descriptors */
	*(__u32 *)buf = 1;	/* tag for this format */
	memcpy (buf + 4, fs, USB_DT_ENDPOINT_SIZE);
	if (HIGHSPEED)
		memcpy (buf + 4 + USB_DT_ENDPOINT_SIZE,
			hs, USB_DT_ENDPOINT_SIZE);
	status = write (fd, buf, 4 + USB_DT_ENDPOINT_SIZE
			+ (HIGHSPEED ? USB_DT_ENDPOINT_SIZE : 0));
	if (status < 0) {
		status = -errno;
		fprintf (stderr, "%s config %s error %d (%s)\n",
			label, name, errno, strerror (errno));
		close (fd);
		return status;
	} else if (verbose) {
		unsigned long	id;

		id = pthread_self ();
		fprintf (stderr, "%s start %ld fd %d\n", label, id, fd);
	}
	return fd;
}

#define source_open(name) \
	ep_config(name,__FUNCTION__, &fs_source_desc, &hs_source_desc)

static char *
build_config (char *cp, const struct usb_endpoint_descriptor **ep)
{
	struct usb_config_descriptor *c;
	int i;

	c = (struct usb_config_descriptor *) cp;

	memcpy (cp, &config, config.bLength);
	cp += config.bLength;
	memcpy (cp, &source_sink_intf, source_sink_intf.bLength);
	cp += source_sink_intf.bLength;
	memcpy (cp, &hid_desc, hid_desc.bLength);
	cp += hid_desc.bLength;

	for (i = 0; i < source_sink_intf.bNumEndpoints; i++) {
		memcpy (cp, ep [i], USB_DT_ENDPOINT_SIZE);
		cp += USB_DT_ENDPOINT_SIZE;
	}
	c->wTotalLength = __cpu_to_le16 (cp - (char *) c);
	return cp;
}

static int init_device (void)
{
	char buf[4096], *cp = &buf[0];
	int fd, status;

	status = autoconfig ();
	if (status < 0) {
		//fprintf (stderr, "?? don't recognize /dev/gadget %s device\n",
		//	iso ? "iso" : "bulk");
		return status;
	}

	fd = open (DEVNAME, O_RDWR);
	if (fd < 0) {
		perror (DEVNAME);
		return -errno;
	}

	*(__u32 *)cp = 0;	/* tag for this format */
	cp += 4;

	/* write full then high speed configs */
	cp = build_config (cp, fs_eps);
	if (HIGHSPEED)
		cp = build_config (cp, hs_eps);

	/* and device descriptor at the end */
	memcpy (cp, &device_desc, sizeof device_desc);
	cp += sizeof device_desc;

	status = write (fd, &buf[0], cp - &buf[0]);
	if (status < 0) {
		perror ("write dev descriptors");
		close (fd);
		return status;
	} else if (status != (cp - buf)) {
		fprintf (stderr, "dev init, wrote %d expected %ld\n",
				status, cp - buf);
		close (fd);
		return -EIO;
	}
	return fd;
}

static void handle_control (int fd, struct usb_ctrlrequest *setup)
{
	int		status, tmp;
	__u8		buf [256], outbuf[8];
	__u16		value, index, length;
	int i;
	int data_fd;
	char cFilename[256];

	value = __le16_to_cpu(setup->wValue);
	index = __le16_to_cpu(setup->wIndex);
	length = __le16_to_cpu(setup->wLength);

	if (verbose)
		fprintf(stderr, "SETUP %02x.%02x "
			"v%04x i%04x %d\n",
			setup->bRequestType, setup->bRequest,
			value, index, length);

	/*
	if ((setup->bRequestType & USB_TYPE_MASK) != USB_TYPE_STANDARD)
		goto special;
	*/

	switch (setup->bRequest) {	/* usb 2.0 spec ch9 requests */
		case USB_REQ_GET_DESCRIPTOR:
			switch (value >> 8) {
				case USB_DT_STRING:
					if (setup->bRequestType != USB_DIR_IN)
						goto stall;
					tmp = value & 0x0ff;
					if (verbose > 1)
						fprintf (stderr,
							"STRING get string %d lang %04x\n",
							tmp, index);
					if (tmp != 0 && index != strings.language)
						goto stall;
					status = usb_gadget_get_string (&strings, tmp, buf);
					if (status < 0)
						goto stall;
					tmp = status;
					if (length < tmp)
						tmp = length;
					status = write (fd, buf, tmp);
					if (status < 0) {
						if (errno == EIDRM)
							fprintf (stderr, "string timeout\n");
						else
							perror ("write string data");
					} else if (status != tmp) {
						fprintf (stderr, "short string write, %d\n",
							status);
					}
					break;
				//case 33://USB_DT_HID
				case USB_DT_RPIPE://0x22:
					//goto stall;
			  	tmp = setup->wValue & 0x00ff;
			  	if (verbose)
			  		printf("RPIPE get rpt desc %d, index %04x\n", tmp, setup->wIndex);
			  	if (tmp != 0 || setup->wIndex != 0) goto stall;
			  	status = write(fd, cReportDescriptor1,
			  		hid_desc.desc[setup->wIndex].wDescriptorLength);
			  	if (status < 0){
			  		if (errno == EIDRM)
			     		printf ("GET_REPORT timeout\n");
			     	else
			     		perror ("write GET_REPORT data");
					}
					break;
				default:
					printf("! Ignoring unknown USB_REQ_GET_DESCRIPTOR type %02X\n", value >> 8);
					goto stall;
			}
			return;
		case USB_REQ_SET_CONFIGURATION:
			if (setup->bRequestType == (USB_DIR_OUT | USB_TYPE_CLASS | USB_RECIP_INTERFACE)) {
				// Incoming Odin request
				status = read(fd, buf, length);
				if (status) perror ("Couldn't read incoming request");

				printf("[cmd-%02X", buf[0]);
				for (i = 1; i < length; i++) printf(" %02X", buf[i]);
				printf("] ");

				sprintf(cFilename, "/tmp/gbt-%02X-%02X%02X%02X%02X.bin", buf[0],
					buf[1], buf[2], buf[3], buf[4]);
				data_fd = open(cFilename, O_RDONLY);
				if (data_fd < 0) {
					fprintf(stderr, "Unable to open %s\n", cFilename);
					memset(outbuf, 0, 8);
				} else {
					read(data_fd, outbuf, 8);
					close(data_fd);
				}

				status = 0;
				if (ioctl(retfd, GADGETFS_CLEAR_HALT) < 0) {
					status = errno;
					perror("reset interrupt fd");
				}
				if (status)
					goto stall;


				//fflush(stdout);
				//fflush(stderr);
				//status = write(retfd, outbuf, 0);
				status = write(retfd, outbuf, 8);
				printf("Response: ");
				for (i = 0; i < 8; i++) printf(" %02X", outbuf[i]);
				printf("\n");
				if (status < 0) {
					perror("Writing interrupt response");
				}

				//status = read (fd, &status, 0);

				/*status = ioctl(retfd, GADGETFS_FIFO_FLUSH);
				if (status < 0)
					perror ("fifo flush");

				// ack (a write would stall)
				status = read (fd, &status, 0);
				if (status)
					perror ("ack SET_CONFIGURATION gbt");
				*/

				return;
			}
			if (setup->bRequestType != USB_DIR_OUT)
				goto stall;
			if (verbose)
				fprintf (stderr, "CONFIG #%d\n", value);

			/* Kernel is normally waiting for us to finish reconfiguring
			 * the device.
			 *
			 * Some hardware can't, notably older PXA2xx hardware.  (With
			 * racey and restrictive config change automagic.  PXA 255 is
			 * OK, most PXA 250s aren't.  If it has a UDC CFR register,
			 * it can handle deferred response for SET_CONFIG.)  To handle
			 * such hardware, don't write code this way ... instead, keep
			 * the endpoints always active and don't rely on seeing any
			 * config change events, either this or SET_INTERFACE.
			 */
			switch (value) {
				case CONFIG_VALUE:
					if (retfd < 0) {
						retfd = source_open(EP_IN_NAME);
						if (retfd < 0) {
							perror("Error opening return (interrupt) endpoint");
							break;
						}
						printf("Opened interrupt data out\n");
					}
					break;
				case 0:
					close_fd(&retfd);
					retfd = -1;
					printf("Closed interrupt data out\n");
					break;
				default:
					/* kernel bug -- "can't happen" */
					fprintf (stderr, "? illegal config\n");
					goto stall;
			}

			/* ... ack (a write would stall) */
			status = read (fd, &status, 0);
			if (status)
				perror ("ack SET_CONFIGURATION");
			return;

		case USB_REQ_GET_INTERFACE:
			// Don't know what this request is but the hardware responds with
			// a zero-packet (ack?)
			if (setup->bRequestType == (USB_DIR_OUT | USB_TYPE_CLASS | USB_RECIP_INTERFACE)
					&& index == 0
					&& length == 0) {
				// Hardware returns a zero packet here
				status = read (fd, &status, 0);
				//status = write(fd, &status, 0);
				if (status < 0) {
					//printf("Error %d when writing zero packet", errno);
					perror("ack to cause write zero packet");
				}
				return;
			}

			if (setup->bRequestType != (USB_DIR_IN|USB_RECIP_INTERFACE)
					|| index != 0
					|| length > 1)
				goto stall;

			/* only one altsetting in this driver */
			buf [0] = 0;
			status = write (fd, buf, length);
			if (status < 0) {
				if (errno == EIDRM)
					fprintf (stderr, "GET_INTERFACE timeout\n");
				else
					perror ("write GET_INTERFACE data");
			} else if (status != length) {
				fprintf (stderr, "short GET_INTERFACE write, %d\n",
					status);
			}
			return;

		case USB_REQ_SET_INTERFACE:
			if (setup->bRequestType != USB_RECIP_INTERFACE
					|| index != 0
					|| value != 0)
				goto stall;

			/* just reset toggle/halt for the interface's endpoints */
			status = 0;
			if (ioctl(retfd, GADGETFS_CLEAR_HALT) < 0) {
				status = errno;
				perror("reset interrupt fd");
			}
			if (status)
				goto stall;

			/* ... and ack (a write would stall) */
			status = read (fd, &status, 0);
			if (status)
				perror ("ack SET_INTERFACE");
			return;

		case 0x01:
			//status = read (fd, &status, 0);
			status = write(fd, &status, 0);
			return;

		default:
			goto stall;
	}

stall:
	if (verbose)
		fprintf (stderr, "... protocol stall %02x.%02x\n",
			setup->bRequestType, setup->bRequest);

	/* non-iso endpoints are stalled by issuing an i/o request
	 * in the "wrong" direction.  ep0 is special only because
	 * the direction isn't fixed.
	 */
	if (setup->bRequestType & USB_DIR_IN)
		status = read (fd, &status, 0);
	else
		status = write (fd, &status, 0);
	if (status != -1)
		fprintf (stderr, "can't stall ep0 for %02x.%02x\n",
			setup->bRequestType, setup->bRequest);
	else if (errno != EL2HLT)
		perror ("ep0 stall");
}

static void signothing(int sig, siginfo_t *info, void *ptr)
{
	/* NOP */
	if (verbose > 2)
		fprintf (stderr, "%s %d\n", __FUNCTION__, sig);
}

static const char *speed(enum usb_device_speed s)
{
	switch (s) {
		case USB_SPEED_LOW:  return "low speed";
		case USB_SPEED_FULL: return "full speed";
		case USB_SPEED_HIGH: return "high speed";
		default: return "UNKNOWN speed";
	}
}

#define	NEVENT		5
#define	LOGDELAY	(15 * 60)	/* seconds before stdout timestamp */

static void *ep0_thread (void *param)
{
	int			fd = *(int*) param;
	struct sigaction	action;
	time_t			now, last;
	struct pollfd		ep0_poll;

	ep0 = pthread_self ();
	pthread_cleanup_push (close_fd, param);

	/* REVISIT signal handling ... normally one pthread should
	 * be doing sigwait() to handle all async signals.
	 */
	action.sa_sigaction = signothing;
	sigfillset (&action.sa_mask);
	action.sa_flags = SA_SIGINFO;
	if (sigaction (SIGINT, &action, NULL) < 0) {
		perror ("SIGINT");
		return 0;
	}
	if (sigaction (SIGQUIT, &action, NULL) < 0) {
		perror ("SIGQUIT");
		return 0;
	}

	ep0_poll.fd = fd;
	ep0_poll.events = POLLIN | POLLOUT | POLLHUP;

	/* event loop */
	last = 0;
	for (;;) {
		int				tmp;
		struct usb_gadgetfs_event	event [NEVENT];
		int				connected = 0;
		int				i, nevent;

		/* Use poll() to test that mechanism, to generate
		 * activity timestamps, and to make it easier to
		 * tweak this code to work without pthreads.  When
		 * AIO is needed without pthreads, ep0 can be driven
		 * instead using SIGIO.
		 */
		tmp = poll(&ep0_poll, 1, -1);
		if (verbose) {
			time (&now);
			if ((now - last) > LOGDELAY) {
				char		timebuf[26];

				last = now;
				ctime_r (&now, timebuf);
				printf ("\n** %s", timebuf);
			}
		}
		if (tmp < 0) {
			/* exit path includes EINTR exits */
			perror("poll");
			break;
		}

		tmp = read (fd, &event, sizeof event);
		if (tmp < 0) {
			if (errno == EAGAIN) {
				sleep (1);
				continue;
			}
			perror ("ep0 read after poll");
			goto done;
		}
		nevent = tmp / sizeof event [0];
		if (nevent != 1 && verbose)
			fprintf (stderr, "read %d ep0 events\n",
				nevent);

		for (i = 0; i < nevent; i++) {
			switch (event [i].type) {
				case GADGETFS_NOP:
					if (verbose)
						fprintf (stderr, "NOP\n");
					break;
				case GADGETFS_CONNECT:
					connected = 1;
					current_speed = event [i].u.speed;
					if (verbose)
						fprintf (stderr, "CONNECT %s\n", speed(event [i].u.speed));
					break;
				case GADGETFS_SETUP:
					connected = 1;
					handle_control(fd, &event [i].u.setup);
					break;
				case GADGETFS_DISCONNECT:
					connected = 0;
					current_speed = USB_SPEED_UNKNOWN;
					if (verbose)
						fprintf(stderr, "DISCONNECT\n");
					break;
				case GADGETFS_SUSPEND:
					// connected = 1;
					if (verbose)
						fprintf (stderr, "SUSPEND\n");
					break;
				default:
					fprintf(stderr, "* unhandled event %d\n", event[i].type);
			}
		}
		continue;
done:
		fflush (stdout);
		//if (connected)
			//stop_io ();
		break;
	}
	if (verbose)
		fprintf (stderr, "done\n");
	fflush (stdout);

	pthread_cleanup_pop (1);
	return 0;
}

int main (int argc, char **argv)
{
	int fd;

	if (chdir ("/dev/gadget") < 0) {
		perror ("can't chdir /dev/gadget");
		return 1;
	}

	fd = init_device ();
	if (fd < 0)
		return 1;
	fprintf (stderr, "/dev/gadget/%s ep0 configured\n", DEVNAME);
	fflush (stderr);

	/*retfd = source_open(EP_IN_NAME);
	if (retfd < 0) {
		perror("Error opening return (interrupt) endpoint");
		return 1;//break;
	}
	printf("Opened interrupt data out\n");*/

	(void) ep0_thread (&fd);

	if (retfd > 0) close_fd(&retfd);
	return 0;
}
