/*
 * gbt-odin.c - Linux kernel module for USB component of Gigabyte Technology's
 *              "Odin GT" ATX power supplies
 *
 * Copyright (C) 2009 Adam Nielsen <a.nielsen@shikadi.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 */

#include <linux/hid.h>
#include <linux/usb.h>

// Not sure what's meant to go here, none of these work...
//#include "usbhid/usbhid.h"
#include "/usr/src/linux/drivers/hid/usbhid/usbhid.h"
//#include "usbhid.h"

#include <linux/module.h>
#include <linux/jiffies.h>
#include <linux/hwmon.h>
#include <linux/sysfs.h>
#include <linux/hwmon-sysfs.h>
#include <linux/err.h>
#include <linux/leds.h>

#define DRVNAME	"gbt-odin"

#define DRVMSG	DRVNAME ": "

static int set_my_pc_on_fire;
module_param(set_my_pc_on_fire, bool, 0);
MODULE_PARM_DESC(set_my_pc_on_fire,
	"Set to 1 to allow the power supply's voltages to be adjusted.");

#define USB_VENDOR_ID_CHU_YUEN  0x1044
#define USB_DEVICE_ID_ODIN_PSU  0x4001

static const struct hid_device_id odin_devices[] = {
	{ HID_USB_DEVICE(USB_VENDOR_ID_CHU_YUEN, USB_DEVICE_ID_ODIN_PSU) },
	{ }
};
MODULE_DEVICE_TABLE(hid, odin_devices);

struct odin_psu_device {
	struct hid_device *hdev;
	struct hid_report *report;
	struct hid_report *report_in;

	struct led_classdev led_cdev_fan;
	struct device *hwmon_dev;

	unsigned long nextupdate;

	/* Cached sensor values */
	long temp[5];
	long voltage[8];
	long current2[8];
	long fanrpm[2];
};

// TEMP: this is in hid-core.c
int usbhid_wait_io(struct hid_device *hid)
{
        struct usbhid_device *usbhid = hid->driver_data;

        if (!wait_event_timeout(usbhid->wait,
                                (!test_bit(HID_CTRL_RUNNING, &usbhid->iofl) &&
                                !test_bit(HID_OUT_RUNNING, &usbhid->iofl)),
                                        10*HZ)) {
                dbg_hid("timeout waiting for ctrl or out queue to clear\n");
                return -1;
        }

        return 0;
}
// END TEMP

/* Common code to interface with the device */

static void odin_get_data(struct odin_psu_device *odin_psu, int msg)
{
	odin_psu->report->field[0]->value[0] = msg;
	odin_psu->report->field[0]->value[1] = 0x01;
	odin_psu->report->field[0]->value[2] = 0x01;
	odin_psu->report->field[0]->value[3] = 0x01;
	odin_psu->report->field[0]->value[4] = 0x01;

	usbhid_submit_report(odin_psu->hdev, odin_psu->report, USB_DIR_OUT);
	usbhid_wait_io(odin_psu->hdev);

	usbhid_submit_report(odin_psu->hdev, odin_psu->report_in, USB_DIR_IN);
	usbhid_wait_io(odin_psu->hdev);
	usbhid_wait_io(odin_psu->hdev);

	// TODO: Sometimes this doesn't wait long enough??
}

/* Code for "led" class */

static void odin_led_fan_set(struct led_classdev *led_cdev,
	enum led_brightness value)
{
	struct odin_psu_device *odin_psu =
		container_of(led_cdev, struct odin_psu_device, led_cdev_fan);

	odin_psu->report->field[0]->value[0] = 0x0b;
	odin_psu->report->field[0]->value[1] = value ? 0x00 : 0x01;
	odin_psu->report->field[0]->value[2] = 0x00;
	odin_psu->report->field[0]->value[3] = 0x00;
	odin_psu->report->field[0]->value[4] = 0x00;

	usbhid_submit_report(odin_psu->hdev, odin_psu->report, USB_DIR_OUT);
	return;
}

/* Code for "hwmon" class */

static char *odin_hwmon_rail_labels[] = {
	"3.3V\n",
	"5V\n",
	"5VSB\n",
	"12V1\n",
	"12V2\n",
	"12V3\n",
	"12V4\n",
	"-12V\n",
	"Total\n"
};

static char *odin_hwmon_labels[] = {
	//"Gigabyte Odin GT power supply\n",
	"odin\n",
	"Internal\n",
	"External 1\n",
	"External 2\n",
	"External 3\n",
	"External 4\n",
	"PSU Fan\n",
	"System Fan\n"
};

static ssize_t odin_hwmon_get_label(struct device *dev, struct device_attribute *devattr, char *buf)
{
	struct sensor_device_attribute *attr = to_sensor_dev_attr(devattr);
	int len = strlen(odin_hwmon_labels[attr->index]);
	memcpy(buf, odin_hwmon_labels[attr->index], len + 1);
	return len;
}

static ssize_t odin_hwmon_get_rail_label(struct device *dev, struct device_attribute *devattr, char *buf)
{
	struct sensor_device_attribute *attr = to_sensor_dev_attr(devattr);
	int len;
	//if (attr->index >= sizeof(odin_hwmon_labels) / sizeof(char *)) return -ENODEV;
	len = strlen(odin_hwmon_rail_labels[attr->index]);
	memcpy(buf, odin_hwmon_rail_labels[attr->index], len);
	return len;
}

/* Get a 16-bit value out of two bytes in big-endian order */
#define get_uint16be_reg(report, offset) \
	((report[(offset % 4) * 2] << 8) | report[(offset % 4) * 2 + 1])

static inline long TEMP_FROM_REG(u8 reg)
{
	return (reg * 1000);
}

static inline long MILLIVOLTS_FROM_REG_5V(u16 reg)
{
	return reg / 10.097;
}

static inline long MILLIVOLTS_FROM_REG_12V(u16 reg)
{
	return reg / 4.036;
}

static inline long MILLIVOLTS_FROM_REG_MINUS12V(u16 reg)
{
	return (((float)reg / 1834.303627 - 32.485) * 1000); // TODO: optimise
}

static inline long RPM_FROM_REG(u16 reg)
{
	return 600000 / (reg * 109.225 / 256);
}

/* The order here must match the Odin register layout */
double odin_reg2amp_div[] = {
	8,     // 5V
	8.38,  // 3.3V
	8,     // unknown / guessing formula (5VSB?)
	11.1,  // 12V1
	11.5,  // 12V2
	9.9,   // 12V3
	11.1,  // unknown / guessing formula (-12V?)
	6.87,  // 12V4
};
double odin_reg2amp_offset[] = {
	-0.37,  // 5V
	0,      // 3.3V
	-0.37,  // unknown / guessing
	-0.23,  // 12V1
	-0.17,  // 12V2
	-0.25,  // 12V3
	-0.25,  // unknown / guessing
	-0.6,   // 12V4
};

static inline long MILLIAMPS_FROM_REG(u8 reg, u8 index)
{
	long ma = ((double)reg / odin_reg2amp_div[index] +
		odin_reg2amp_offset[index]) * 1000.0;
	return (ma > 0) ? ma : 0;
}

static struct odin_psu_device *odin_update(struct device *dev)
{
	struct odin_psu_device *odin_psu = dev_get_drvdata(dev);
	__s32 *reg = odin_psu->report_in->field[0]->value;

	/* Don't update again if we only just have done */
	if (time_is_after_eq_jiffies(odin_psu->nextupdate)) return odin_psu;

	odin_get_data(odin_psu, 0x02);
	odin_psu->current2[1] = MILLIAMPS_FROM_REG(reg[0], 0);
	odin_psu->current2[0] = MILLIAMPS_FROM_REG(reg[1], 1);
	odin_psu->current2[2] = MILLIAMPS_FROM_REG(reg[2], 2);
	odin_psu->current2[3] = MILLIAMPS_FROM_REG(reg[3], 3);
	odin_psu->current2[4] = MILLIAMPS_FROM_REG(reg[4], 4);
	odin_psu->current2[5] = MILLIAMPS_FROM_REG(reg[5], 5);
	odin_psu->current2[7] = MILLIAMPS_FROM_REG(reg[6], 6);
	odin_psu->current2[6] = MILLIAMPS_FROM_REG(reg[7], 7);

	odin_get_data(odin_psu, 0x03);
	odin_psu->fanrpm[0] = RPM_FROM_REG(get_uint16be_reg(reg, 0));
	odin_psu->fanrpm[1] = RPM_FROM_REG(get_uint16be_reg(reg, 1));

	odin_get_data(odin_psu, 0x04);
	odin_psu->temp[1] = TEMP_FROM_REG(reg[0]);
	odin_psu->temp[2] = TEMP_FROM_REG(reg[2]);
	odin_psu->temp[3] = TEMP_FROM_REG(reg[4]);
	odin_psu->temp[4] = TEMP_FROM_REG(reg[6]);

	odin_get_data(odin_psu, 0x05);
	odin_psu->temp[0] = TEMP_FROM_REG(reg[0]);

	odin_get_data(odin_psu, 0x21);
	odin_psu->voltage[1] = MILLIVOLTS_FROM_REG_5V(get_uint16be_reg(reg, 0));
	odin_psu->voltage[0] = MILLIVOLTS_FROM_REG_5V(get_uint16be_reg(reg, 1));
	odin_psu->voltage[2] = MILLIVOLTS_FROM_REG_5V(get_uint16be_reg(reg, 2));
	odin_psu->voltage[6] = MILLIVOLTS_FROM_REG_12V(get_uint16be_reg(reg, 3));

	odin_get_data(odin_psu, 0x22);
	odin_psu->voltage[3] = MILLIVOLTS_FROM_REG_12V(get_uint16be_reg(reg, 0));
	odin_psu->voltage[4] = MILLIVOLTS_FROM_REG_12V(get_uint16be_reg(reg, 1));
	odin_psu->voltage[5] = MILLIVOLTS_FROM_REG_12V(get_uint16be_reg(reg, 2));
	odin_psu->voltage[7] = MILLIVOLTS_FROM_REG_MINUS12V(get_uint16be_reg(reg, 3));

	/* Don't update cached data for another ~1 second or so */
	odin_psu->nextupdate = jiffies + HZ;

	return odin_psu;
}

static ssize_t odin_hwmon_show_temp(struct device *dev, struct device_attribute *devattr, char *buf)
{
	struct sensor_device_attribute *attr = to_sensor_dev_attr(devattr);
	struct odin_psu_device *odin_psu = odin_update(dev);
	sprintf(buf, "%ld\n", odin_psu->temp[attr->index]);
	return strlen(buf);
}

static ssize_t odin_hwmon_show_fanrpm(struct device *dev, struct device_attribute *devattr, char *buf)
{
	struct sensor_device_attribute *attr = to_sensor_dev_attr(devattr);
	struct odin_psu_device *odin_psu = odin_update(dev);
	sprintf(buf, "%ld\n", odin_psu->fanrpm[attr->index]);
	return strlen(buf);
}

static ssize_t odin_hwmon_show_voltage(struct device *dev, struct device_attribute *devattr, char *buf)
{
	struct sensor_device_attribute *attr = to_sensor_dev_attr(devattr);
	struct odin_psu_device *odin_psu = odin_update(dev);
	sprintf(buf, "%ld\n", odin_psu->voltage[attr->index]);
	return strlen(buf);
}

static ssize_t odin_hwmon_show_current(struct device *dev, struct device_attribute *devattr, char *buf)
{
	struct sensor_device_attribute *attr = to_sensor_dev_attr(devattr);
	struct odin_psu_device *odin_psu = odin_update(dev);
	sprintf(buf, "%ld\n", odin_psu->current2[attr->index]);
	return strlen(buf);
}

static ssize_t odin_hwmon_show_power(struct device *dev, struct device_attribute *devattr, char *buf)
{
	struct sensor_device_attribute *attr = to_sensor_dev_attr(devattr);
	struct odin_psu_device *odin_psu = odin_update(dev);
	int i;

	/* milliamp * millivolt == microwatt */
	long microwatts = 0; /* keep compiler happy */
	if (attr->index == 8) {
		/* Total power */
		for (i = 0; i < 8; i++) {
			if ((i == 2) || (i == 7)) continue; // ignore unknown values
			microwatts += odin_psu->current2[i] * odin_psu->voltage[i];
		}
	} else {
		microwatts = odin_psu->current2[attr->index] * odin_psu->voltage[attr->index];
	}

	sprintf(buf, "%ld\n", microwatts);
	return strlen(buf);
}

/* All the different hwmon values we provide.  Some are commented out because
 * there are values returned but we don't know what they're for.
 */
static SENSOR_DEVICE_ATTR(name, S_IRUGO, odin_hwmon_get_label, NULL, 0);

static SENSOR_DEVICE_ATTR(temp1_label, S_IRUGO, odin_hwmon_get_label, NULL, 1);
static SENSOR_DEVICE_ATTR(temp2_label, S_IRUGO, odin_hwmon_get_label, NULL, 2);
static SENSOR_DEVICE_ATTR(temp3_label, S_IRUGO, odin_hwmon_get_label, NULL, 3);
static SENSOR_DEVICE_ATTR(temp4_label, S_IRUGO, odin_hwmon_get_label, NULL, 4);
static SENSOR_DEVICE_ATTR(temp5_label, S_IRUGO, odin_hwmon_get_label, NULL, 5);

static SENSOR_DEVICE_ATTR(fan1_label, S_IRUGO, odin_hwmon_get_label, NULL, 6);
static SENSOR_DEVICE_ATTR(fan2_label, S_IRUGO, odin_hwmon_get_label, NULL, 7);

static SENSOR_DEVICE_ATTR(in0_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 0);
static SENSOR_DEVICE_ATTR(in1_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 1);
static SENSOR_DEVICE_ATTR(in2_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 2);
static SENSOR_DEVICE_ATTR(in3_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 3);
static SENSOR_DEVICE_ATTR(in4_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 4);
static SENSOR_DEVICE_ATTR(in5_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 5);
static SENSOR_DEVICE_ATTR(in6_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 6);
static SENSOR_DEVICE_ATTR(in7_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 7);

static SENSOR_DEVICE_ATTR(curr0_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 0);
static SENSOR_DEVICE_ATTR(curr1_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 1);
static SENSOR_DEVICE_ATTR(curr2_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 2);
static SENSOR_DEVICE_ATTR(curr3_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 3);
static SENSOR_DEVICE_ATTR(curr4_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 4);
static SENSOR_DEVICE_ATTR(curr5_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 5);
static SENSOR_DEVICE_ATTR(curr6_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 6);
static SENSOR_DEVICE_ATTR(curr7_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 7);

static SENSOR_DEVICE_ATTR(power0_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 0);
static SENSOR_DEVICE_ATTR(power1_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 1);
//static SENSOR_DEVICE_ATTR(power2_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 2);
static SENSOR_DEVICE_ATTR(power3_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 3);
static SENSOR_DEVICE_ATTR(power4_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 4);
static SENSOR_DEVICE_ATTR(power5_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 5);
static SENSOR_DEVICE_ATTR(power6_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 6);
//static SENSOR_DEVICE_ATTR(power7_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 7);
static SENSOR_DEVICE_ATTR(power8_label, S_IRUGO, odin_hwmon_get_rail_label, NULL, 8);

static SENSOR_DEVICE_ATTR(temp1_input, S_IRUGO, odin_hwmon_show_temp, NULL, 0);
static SENSOR_DEVICE_ATTR(temp2_input, S_IRUGO, odin_hwmon_show_temp, NULL, 1);
static SENSOR_DEVICE_ATTR(temp3_input, S_IRUGO, odin_hwmon_show_temp, NULL, 2);
static SENSOR_DEVICE_ATTR(temp4_input, S_IRUGO, odin_hwmon_show_temp, NULL, 3);
static SENSOR_DEVICE_ATTR(temp5_input, S_IRUGO, odin_hwmon_show_temp, NULL, 4);

static SENSOR_DEVICE_ATTR(fan1_input, S_IRUGO, odin_hwmon_show_fanrpm, NULL, 0);
static SENSOR_DEVICE_ATTR(fan2_input, S_IRUGO, odin_hwmon_show_fanrpm, NULL, 1);

static SENSOR_DEVICE_ATTR(in0_input, S_IRUGO, odin_hwmon_show_voltage, NULL, 0);
static SENSOR_DEVICE_ATTR(in1_input, S_IRUGO, odin_hwmon_show_voltage, NULL, 1);
static SENSOR_DEVICE_ATTR(in2_input, S_IRUGO, odin_hwmon_show_voltage, NULL, 2);
static SENSOR_DEVICE_ATTR(in3_input, S_IRUGO, odin_hwmon_show_voltage, NULL, 3);
static SENSOR_DEVICE_ATTR(in4_input, S_IRUGO, odin_hwmon_show_voltage, NULL, 4);
static SENSOR_DEVICE_ATTR(in5_input, S_IRUGO, odin_hwmon_show_voltage, NULL, 5);
static SENSOR_DEVICE_ATTR(in6_input, S_IRUGO, odin_hwmon_show_voltage, NULL, 6);
static SENSOR_DEVICE_ATTR(in7_input, S_IRUGO, odin_hwmon_show_voltage, NULL, 7);

static SENSOR_DEVICE_ATTR(curr0_input, S_IRUGO, odin_hwmon_show_current, NULL, 0);
static SENSOR_DEVICE_ATTR(curr1_input, S_IRUGO, odin_hwmon_show_current, NULL, 1);
static SENSOR_DEVICE_ATTR(curr2_input, S_IRUGO, odin_hwmon_show_current, NULL, 2);
static SENSOR_DEVICE_ATTR(curr3_input, S_IRUGO, odin_hwmon_show_current, NULL, 3);
static SENSOR_DEVICE_ATTR(curr4_input, S_IRUGO, odin_hwmon_show_current, NULL, 4);
static SENSOR_DEVICE_ATTR(curr5_input, S_IRUGO, odin_hwmon_show_current, NULL, 5);
static SENSOR_DEVICE_ATTR(curr6_input, S_IRUGO, odin_hwmon_show_current, NULL, 6);
static SENSOR_DEVICE_ATTR(curr7_input, S_IRUGO, odin_hwmon_show_current, NULL, 7);

static SENSOR_DEVICE_ATTR(power0_input, S_IRUGO, odin_hwmon_show_power, NULL, 0);
static SENSOR_DEVICE_ATTR(power1_input, S_IRUGO, odin_hwmon_show_power, NULL, 1);
//static SENSOR_DEVICE_ATTR(power2_input, S_IRUGO, odin_hwmon_show_power, NULL, 2);
static SENSOR_DEVICE_ATTR(power3_input, S_IRUGO, odin_hwmon_show_power, NULL, 3);
static SENSOR_DEVICE_ATTR(power4_input, S_IRUGO, odin_hwmon_show_power, NULL, 4);
static SENSOR_DEVICE_ATTR(power5_input, S_IRUGO, odin_hwmon_show_power, NULL, 5);
static SENSOR_DEVICE_ATTR(power6_input, S_IRUGO, odin_hwmon_show_power, NULL, 6);
//static SENSOR_DEVICE_ATTR(power7_input, S_IRUGO, odin_hwmon_show_power, NULL, 7);
static SENSOR_DEVICE_ATTR(power8_input, S_IRUGO, odin_hwmon_show_power, NULL, 8);

static struct attribute *odin_attributes[] = {
	&sensor_dev_attr_name.dev_attr.attr,

	&sensor_dev_attr_temp1_label.dev_attr.attr,
	&sensor_dev_attr_temp2_label.dev_attr.attr,
	&sensor_dev_attr_temp3_label.dev_attr.attr,
	&sensor_dev_attr_temp4_label.dev_attr.attr,
	&sensor_dev_attr_temp5_label.dev_attr.attr,

	&sensor_dev_attr_fan1_label.dev_attr.attr,
	&sensor_dev_attr_fan2_label.dev_attr.attr,

	&sensor_dev_attr_in0_label.dev_attr.attr,
	&sensor_dev_attr_in1_label.dev_attr.attr,
	&sensor_dev_attr_in2_label.dev_attr.attr,
	&sensor_dev_attr_in3_label.dev_attr.attr,
	&sensor_dev_attr_in4_label.dev_attr.attr,
	&sensor_dev_attr_in5_label.dev_attr.attr,
	&sensor_dev_attr_in6_label.dev_attr.attr,
	&sensor_dev_attr_in7_label.dev_attr.attr,

	&sensor_dev_attr_curr0_label.dev_attr.attr,
	&sensor_dev_attr_curr1_label.dev_attr.attr,
	&sensor_dev_attr_curr2_label.dev_attr.attr,
	&sensor_dev_attr_curr3_label.dev_attr.attr,
	&sensor_dev_attr_curr4_label.dev_attr.attr,
	&sensor_dev_attr_curr5_label.dev_attr.attr,
	&sensor_dev_attr_curr6_label.dev_attr.attr,
	&sensor_dev_attr_curr7_label.dev_attr.attr,

	&sensor_dev_attr_power0_label.dev_attr.attr,
	&sensor_dev_attr_power1_label.dev_attr.attr,
//	&sensor_dev_attr_power2_label.dev_attr.attr,
	&sensor_dev_attr_power3_label.dev_attr.attr,
	&sensor_dev_attr_power4_label.dev_attr.attr,
	&sensor_dev_attr_power5_label.dev_attr.attr,
	&sensor_dev_attr_power6_label.dev_attr.attr,
//	&sensor_dev_attr_power7_label.dev_attr.attr,
	&sensor_dev_attr_power8_label.dev_attr.attr,

	&sensor_dev_attr_temp1_input.dev_attr.attr,
	&sensor_dev_attr_temp2_input.dev_attr.attr,
	&sensor_dev_attr_temp3_input.dev_attr.attr,
	&sensor_dev_attr_temp4_input.dev_attr.attr,
	&sensor_dev_attr_temp5_input.dev_attr.attr,

	&sensor_dev_attr_fan1_input.dev_attr.attr,
	&sensor_dev_attr_fan2_input.dev_attr.attr,

	&sensor_dev_attr_in0_input.dev_attr.attr,
	&sensor_dev_attr_in1_input.dev_attr.attr,
	&sensor_dev_attr_in2_input.dev_attr.attr,
	&sensor_dev_attr_in3_input.dev_attr.attr,
	&sensor_dev_attr_in4_input.dev_attr.attr,
	&sensor_dev_attr_in5_input.dev_attr.attr,
	&sensor_dev_attr_in6_input.dev_attr.attr,
	&sensor_dev_attr_in7_input.dev_attr.attr,

	&sensor_dev_attr_curr0_input.dev_attr.attr,
	&sensor_dev_attr_curr1_input.dev_attr.attr,
	&sensor_dev_attr_curr2_input.dev_attr.attr,
	&sensor_dev_attr_curr3_input.dev_attr.attr,
	&sensor_dev_attr_curr4_input.dev_attr.attr,
	&sensor_dev_attr_curr5_input.dev_attr.attr,
	&sensor_dev_attr_curr6_input.dev_attr.attr,
	&sensor_dev_attr_curr7_input.dev_attr.attr,

	&sensor_dev_attr_power0_input.dev_attr.attr,
	&sensor_dev_attr_power1_input.dev_attr.attr,
//	&sensor_dev_attr_power2_input.dev_attr.attr,
	&sensor_dev_attr_power3_input.dev_attr.attr,
	&sensor_dev_attr_power4_input.dev_attr.attr,
	&sensor_dev_attr_power5_input.dev_attr.attr,
	&sensor_dev_attr_power6_input.dev_attr.attr,
//	&sensor_dev_attr_power7_input.dev_attr.attr,
	&sensor_dev_attr_power8_input.dev_attr.attr,

	NULL
};

static const struct attribute_group odin_attr_group = {
	.attrs = odin_attributes,
};

/* This function adds devices to the system for interfacing with the PSU.  It
   is called from odin_probe() when new devices are plugged into the system. */
static int odin_add_devices(struct hid_device *hdev)
{
	int err;
	struct odin_psu_device *odin_psu;
	struct hid_report *report;
	struct hid_report *report_in;
	struct list_head *report_list = &hdev->report_enum[HID_OUTPUT_REPORT].report_list;
	struct list_head *report_in_list = &hdev->report_enum[HID_INPUT_REPORT].report_list;

	/* Make sure all the HID stuff is as we expect */

	if (list_empty(report_list)) {
		dev_err(&hdev->dev, "no output report found\n");
		return -ENODEV;
	}

	if (list_empty(report_in_list)) {
		dev_err(&hdev->dev, "no input report found\n");
		return -ENODEV;
	}

	report = list_entry(report_list->next, struct hid_report, list);

	if (report->maxfield < 1) {
		dev_err(&hdev->dev, "not enough fields in report\n");
		return -ENODEV;
	}

	report_in = list_entry(report_in_list->next, struct hid_report, list);
	if (report_in->maxfield < 1) {
		dev_err(&hdev->dev, "not enough fields in input report\n");
		return -ENODEV;
	}

	/* HID looks good, get ourselves up and running */

	odin_psu = kzalloc(sizeof(struct odin_psu_device), GFP_KERNEL);
	if (!odin_psu) return -ENOMEM;

	odin_psu->hdev = hdev;
	odin_psu->report = report;
	odin_psu->report_in = report_in;
	odin_psu->nextupdate = jiffies;

	/* Attach our dynamically allocated data to the "hid" device */
	hid_set_drvdata(hdev, odin_psu);

	/* Register a device in the "led" class to control the LED built into
	   the power supply fan. */
	odin_psu->led_cdev_fan.name = DRVNAME ":blue:fan";
	odin_psu->led_cdev_fan.brightness_set = odin_led_fan_set;
	if ((err = led_classdev_register(&hdev->dev, &odin_psu->led_cdev_fan)))
		goto exit_free_psu_mem;

	/* Register a device in the "hwmon" class to provide sensor readings */
	odin_psu->hwmon_dev = hwmon_device_register(&hdev->dev);
	if (IS_ERR(odin_psu->hwmon_dev)) {
		err = PTR_ERR(odin_psu->hwmon_dev);
		dev_err(&hdev->dev, "Class registration failed (%d)\n", err);
		goto exit_hwmon_class;
	}
	dev_set_drvdata(odin_psu->hwmon_dev, odin_psu);

	//if ((err = sysfs_create_group(&odin_psu->hwmon_dev->kobj, &odin_attr_group))) {
	if ((err = sysfs_create_group(&odin_psu->hdev->dev.kobj, &odin_attr_group))) {
		dev_err(&hdev->dev, "Unable to create sysfs entries (%d)\n", err);
		goto exit_led_class;
	}

	dev_info(&hdev->dev, "Hardware monitoring for Gigabyte Odin PSUs by "
		"Adam Nielsen <a.nielsen@shikadi.net>\n");

	return 0;

exit_hwmon_class:
	//sysfs_remove_group(&odin_psu->hwmon_dev->kobj, &odin_attr_group);
	sysfs_remove_group(&odin_psu->hdev->dev.kobj, &odin_attr_group);
exit_led_class:
	led_classdev_unregister(&odin_psu->led_cdev_fan);
exit_free_psu_mem:
	kfree(odin_psu);
	return err;
}

static int odin_probe(struct hid_device *hdev, const struct hid_device_id *id)
{
	int err;

	dev_info(&hdev->dev, "Attaching to device\n");

	/* No device set up is required, so call hid_parse() right away. */
	if ((err = hid_parse(hdev))) {
		dev_err(&hdev->dev, "hid_parse() failed\n");
		goto exit;
	}

	/* Start the underlying hardware.  Once this is successful, hid_hw_stop()
	 * must eventually be called.
	 */
	if ((err = hid_hw_start(hdev, HID_CONNECT_DEFAULT))) {
		dev_err(&hdev->dev, "hid_hw_start() failed\n");
		goto exit;
	}

	/* Create devices for interfacing with the PSU */
	if ((err = odin_add_devices(hdev))) goto exit_hw_start;

	return 0;

exit_hw_start:
	hid_hw_stop(hdev);
exit:
	return err;
}

static void odin_remove(struct hid_device *hdev)
{
	struct odin_psu_device *odin_psu = hid_get_drvdata(hdev);

	dev_info(&hdev->dev, "Disconnecting from device\n");

	/* Remove the "hwmon" class device */
	//sysfs_remove_group(&odin_psu->hwmon_dev->kobj, &odin_attr_group);
	sysfs_remove_group(&odin_psu->hdev->dev.kobj, &odin_attr_group);
	hwmon_device_unregister(odin_psu->hwmon_dev);
	led_classdev_unregister(&odin_psu->led_cdev_fan);
	hid_hw_stop(hdev);

	/* Detach our data from the driver and free it */
	hid_set_drvdata(hdev, NULL);
	kfree(odin_psu);

	return;
}

static struct hid_driver odin_driver = {
	.name = DRVNAME,
	.id_table = odin_devices,
	.probe = odin_probe,
	.remove = odin_remove
	// TODO: suspend+resume? see docs/leds-fsg.c
};

static int odin_init(void)
{
	/* This message is important as it will be the only notification anything
	 * has happened if some other driver has grabbed the device.
	 */
	printk(KERN_NOTICE DRVMSG "Registering with HID subsystem\n");
	return hid_register_driver(&odin_driver);
}

static void odin_exit(void)
{
	hid_unregister_driver(&odin_driver);
	return;
}

module_init(odin_init);
module_exit(odin_exit);

MODULE_AUTHOR("Adam Nielsen <a.nielsen@shikadi.net>");
MODULE_DESCRIPTION("Gigabyte Odin PSU sensor monitor");
MODULE_LICENSE("GPL");
